from urllib.parse import urlparse

import requests
from loguru import logger
from opentelemetry import context, trace
from opentelemetry.instrumentation.requests import RequestsInstrumentor
from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator
from wrapt import wrap_function_wrapper


class CustomRequestsInstrumentor(RequestsInstrumentor):
    def __init__(self, tracer=None):
        super().__init__()
        self._original_send = None
        self._propagator = TraceContextTextMapPropagator()
        self._tracer = tracer or trace.get_tracer(__name__)

    def _instrument(self, **kwargs):
        super()._instrument(**kwargs)

        def instrumented_send(wrapped, instance, args, kwargs):
            # Get the current context
            current_context = context.get_current()
            current_span = trace.get_current_span(current_context)

            logger.info(f"\n\n **  Checking current span {current_span}")

            # Extract request from args or kwargs
            request = args[0] if args else kwargs.get('request')

            if not request:
                return wrapped(*args, **kwargs)

            # Check if the URL should be ignored
            parsed_url = urlparse(request.url)
            if parsed_url.path.startswith('/v1/traces'):
                return wrapped(*args, **kwargs)

            tracer = trace.get_tracer(__name__)
            with self._tracer.start_as_current_span(
                f"HTTP {request.method}",
                context=current_context,
                kind=trace.SpanKind.CLIENT,
                attributes={
                    "http.method": request.method,
                    "http.url": request.url,
                },
            ) as span:
                # Inject the current context into the headers
                self._propagator.inject(request.headers, context=current_context)

                # Add custom trace IDs
                # span_context = span.get_span_context()
                # if span_context.is_valid:
                #     trace_id = format(span_context.trace_id, '032x')
                #     span_id = format(span_context.span_id, '016x')
                #     request.headers['X-acuvity-Trace-ID'] = trace_id
                #     request.headers['X-acuvity-Span-ID'] = span_id
                # Add custom trace IDs
                #if span.is_recording():
                span_context = span.get_span_context()
                trace_id = format(span_context.trace_id, '032x')
                span_id = format(span_context.span_id, '016x')
                request.headers['X-acuvity-Trace-ID'] = trace_id
                request.headers['X-acuvity-Span-ID'] = span_id


                logger.info(f"Outgoing request headers: {request.headers}")

                try:
                    response = wrapped(*args, **kwargs)
                    span.set_attribute("http.status_code", response.status_code)
                    return response
                except Exception as e:
                    span.record_exception(e)
                    raise

        wrap_function_wrapper('requests.sessions', 'Session.send', instrumented_send)

    def _uninstrument(self, **kwargs):
        pass  # The original _uninstrument method should handle this



from typing import Optional
from uuid import UUID

import httpx
from loguru import logger
from opentelemetry import context, trace
from opentelemetry.instrumentation.httpx import HTTPXClientInstrumentor
from opentelemetry.trace import Span
from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator


class CustomHTTPXClientInstrumentor(HTTPXClientInstrumentor):
    def __init__(self, tracer=None):
        super().__init__()
        self._original_send = None
        self._propagator = TraceContextTextMapPropagator()
        self._tracer = tracer or trace.get_tracer(__name__)

    def _instrument(self, **kwargs):
        super()._instrument(**kwargs)
        self._original_send = httpx.Client.send

        def instrumented_send(client, request, **kwargs):
            current_context = context.get_current()
            current_span = trace.get_current_span(current_context)
            logger.debug(f"\n Request is {request} and \n args {kwargs}\n")
            tracer = trace.get_tracer(__name__)
            logger.debug(f"\n tracer {self._tracer}, Current CONTEXT is  --> {current_span} \n")
            with self._tracer.start_as_current_span(
                f"HTTP {request.method}",
                context=trace.set_span_in_context(current_span),
                kind=trace.SpanKind.CLIENT,
                attributes={
                    "http.method": request.method,
                    "http.url": str(request.url),
                },
            ) as span:
                # Inject the current context into the headers
                self._propagator.inject(request.headers)

                # Add custom trace IDs
                if span.is_recording():
                    span_context = span.get_span_context()
                    trace_id = format(span_context.trace_id, '032x')
                    span_id = format(span_context.span_id, '016x')
                    request.headers['X-acuvity-Trace-ID'] = trace_id
                    request.headers['X-acuvity-Span-ID'] = span_id

                # Log the headers for inspection
                logger.debug(f"Outgoing request headers: {request.headers} \n")

                try:
                    response = self._original_send(client, request, **kwargs)
                    logger.debug(f"\n response {response}")
                    span.set_attribute("http.status_code", response.status_code)
                    return response
                except Exception as e:
                    span.record_exception(e)
                    raise
            # current_context = trace.get_current_span().get_span_context()
            # logger.debug"\n Current context is  -->", current_context)
            # # Inject the current context into the headers
            # self._propagator.inject(request.headers)

            # # Add custom trace IDs
            # if current_context.is_valid:
            #     trace_id = format(current_context.trace_id, '032x')
            #     span_id = format(current_context.span_id, '016x')
            #     request.headers['X-acuvity-Trace-ID'] = trace_id
            #     request.headers['X-acuvity-Span-ID'] = span_id

            # # Log the headers for inspection
            # logger.debug"CustomHTTPXClientInstrumentor Outgoing request headers:", request.headers)

            # return self._original_send(client, request, **kwargs)

        httpx.Client.send = instrumented_send

    def _uninstrument(self, **kwargs):
        if self._original_send:
            httpx.Client.send = self._original_send
            self._original_send = None

    def get_current_span(self) -> Optional[Span]:
        import langchain_core

        run_id: Optional[UUID] = None
        config = langchain_core.runnables.config.var_child_runnable_config.get()
        logger.warning(f"CONIFG --> {config}")
        if not isinstance(config, dict):
            return None
        for v in config.values():
            if not isinstance(v, langchain_core.callbacks.BaseCallbackManager):
                continue
            if run_id := v.parent_run_id:
                break
        if not run_id:
            return None
        return self._tracer.get_span(run_id)



from opentelemetry.exporter.jaeger.thrift import JaegerExporter
from opentelemetry.sdk.trace.export import BatchSpanProcessor


class CustomJaegerExporter:
    def __init__(self):
        # create a JaegerExporter
        self.jaeger_exporter = JaegerExporter(
            # configure agent
            agent_host_name='localhost',
            agent_port=6831,
            # optional: configure also collector
            # collector_endpoint='http://localhost:14268/api/traces?format=jaeger.thrift',
            # username=xxxx, # optional
            # password=xxxx, # optional
            # max_tag_value_length=None # optional
            udp_split_oversized_batches=True
        )
        #self.exporter = OTLPSpanExporter(endpoint=otlp_endpoint)

    def get_processor(self, max_export_batch_size=512, export_timeout_millis=30000, max_queue_size=2048):
        # Create a BatchSpanProcessor and add the exporter to it
        #span_processor = BatchSpanProcessor(self.jaeger_exporter)
        return BatchSpanProcessor(
            self.jaeger_exporter,
            max_export_batch_size=max_export_batch_size,
            export_timeout_millis=export_timeout_millis,
            max_queue_size=max_queue_size
        )



from typing import TYPE_CHECKING, Any, Callable, Collection, Optional
from uuid import UUID

import langchain_core
from openinference.instrumentation import OITracer, TraceConfig
from opentelemetry import trace as trace_api
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor  # type: ignore
from opentelemetry.trace import Span
from wrapt import wrap_function_wrapper  # type: ignore

from instrumentors.python.langchain._tracer import OpenInferenceTracer
from instrumentors.python.langchain.package import _instruments
from instrumentors.python.langchain.version import __version__

if TYPE_CHECKING:
    from langchain_core.callbacks import BaseCallbackManager
    from langchain_core.runnables.config import var_child_runnable_config  # noqa F401
    from openinference.instrumentation.langchain._tracer import OpenInferenceTracer

from loguru import logger

from instrumentors.python.langchain.httpx_instrumentor import (
    CustomHTTPXClientInstrumentor,
)
from instrumentors.python.langchain.requests_instrumentor import (
    CustomRequestsInstrumentor,
)


class LangChainInstrumentor(BaseInstrumentor):  # type: ignore
    """
    An instrumentor for LangChain
    """

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs: Any) -> None:
        if not (tracer_provider := kwargs.get("tracer_provider")):
            tracer_provider = trace_api.get_tracer_provider()
        if not (config := kwargs.get("config")):
            config = TraceConfig()
        else:
            assert isinstance(config, TraceConfig)


        logger.info(f"\n **** ABHI TRACE provider : {tracer_provider}")
        tracer = OITracer(
            trace_api.get_tracer(__name__, __version__, tracer_provider),
            config=config,
        )
        self._tracer: Optional[OpenInferenceTracer] = OpenInferenceTracer(tracer)
        self._requests_instrumentor = CustomRequestsInstrumentor(self._tracer)
        self._requests_instrumentor.instrument(tracer_provider=tracer_provider)
        self._httpx_instrumentor = CustomHTTPXClientInstrumentor(self._tracer)
        self._httpx_instrumentor.instrument(tracer_provider=tracer_provider)


        self._original_callback_manager_init = langchain_core.callbacks.BaseCallbackManager.__init__
        wrap_function_wrapper(
            module="langchain_core.callbacks",
            name="BaseCallbackManager.__init__",
            wrapper=_BaseCallbackManagerInit(self._tracer),
        )

    def _uninstrument(self, **kwargs: Any) -> None:

        langchain_core.callbacks.BaseCallbackManager.__init__ = self._original_callback_manager_init  # type: ignore
        self._original_callback_manager_init = None  # type: ignore
        self._tracer = None
        if self._httpx_instrumentor:
            self._httpx_instrumentor.uninstrument()

    def get_span(self, run_id: UUID) -> Optional[Span]:
        return self._tracer.get_span(run_id) if self._tracer else None


class _BaseCallbackManagerInit:
    __slots__ = ("_tracer",)

    def __init__(self, tracer: "OpenInferenceTracer"):
        self._tracer = tracer

    def __call__(
        self,
        wrapped: Callable[..., None],
        instance: "BaseCallbackManager",
        args: Any,
        kwargs: Any,
    ) -> None:
        wrapped(*args, **kwargs)
        for handler in instance.inheritable_handlers:
            # Handlers may be copied when new managers are created, so we
            # don't want to keep adding. E.g. see the following location.
            # https://github.com/langchain-ai/langchain/blob/5c2538b9f7fb64afed2a918b621d9d8681c7ae32/libs/core/langchain_core/callbacks/manager.py#L1876  # noqa: E501
            if isinstance(handler, type(self._tracer)):
                break
        else:
            instance.add_handler(self._tracer, True)
